print("import: 'pyAgrum'")
import pyAgrum

print("import: 'pyAgrum.lib.bn2roc'")
import pyAgrum.lib.bn2roc

print("import: 'pyAgrum.lib.bn2scores'")
import pyAgrum.lib.bn2scores

print("import: 'pyAgrum.lib.bn_vs_bn'")
import pyAgrum.lib.bn_vs_bn

print("import: 'pyAgrum.lib.dynamicBN'")
import pyAgrum.lib.dynamicBN

print("import: 'pyAgrum.lib.notebook'")
import pyAgrum.lib.notebook

